/*

	Quick and Dirty Game Development Framework (QDGDF)

	Copyright (C) 2001 Angel Ortega <angel@triptico.com>

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

	http://www.triptico.com

*/

#include "qdgdf_audio.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>

#include <linux/soundcard.h>
#include <linux/cdrom.h>


/*******************
	Data
*******************/

/* the audio device */
static int _audio_fd=-1;

/* the pipe */
static int _command_fd=-1;

/* daemon pid */
static int _daemon_pid=-1;

/** commands **/

#define SOUND_CMD_EXIT	  'X'
#define SOUND_CMD_RESET   'R'
#define SOUND_CMD_LOAD	  'L'
#define SOUND_CMD_START   'S'
#define SOUND_CMD_STOP	  'T'
#define SOUND_CMD_SETPAN  'P'
#define SOUND_CMD_PAUSE   'A'
#define SOUND_CMD_UNPAUSE 'U'

/* sequential numbers */
static int _program_seq;
static int _daemon_seq;

/* the sound buffers */

#define SOUND_BUFFER_BITS 10
#define SOUND_BUFFER_SIZE (1<<SOUND_BUFFER_BITS)

static unsigned char _sound_buffer_8[SOUND_BUFFER_SIZE];
static short int _sound_buffer_16[SOUND_BUFFER_SIZE];

/** sounds **/

#define MAX_SOUNDS 64

static struct
{
	void * data;		/* the real data */
	unsigned char * data_8;
	short int * data_16;
	int size;
	int offset;
	int loop;		/* 0, 1 */
	int pan;		/* 1, 0, -1 */

} _sounds[MAX_SOUNDS];


static char msg[256];

/* cdrom */
/* static int _cdrom_fd=-1; */


/********************
	Cdigo
********************/


int detect_endianess(void)
/* is this machine big or little endian? */
{
	short s;
	unsigned char * c;

	s=1;
	c=(unsigned char *) &s;

	if(*c==1) return(AFMT_S16_LE);
	return(AFMT_S16_BE);
}


static int open_device(void)
{
	int n;

	if((_audio_fd=open("/dev/dsp", O_WRONLY))==-1)
		return(0);

	n=(2<<16)|SOUND_BUFFER_BITS;
	if(ioctl(_audio_fd, SNDCTL_DSP_SETFRAGMENT, &n)==-1)
		return(0);

	if(ioctl(_audio_fd, SNDCTL_DSP_RESET, 0)==-1)
		return(0);

	n=11025;
	if(ioctl(_audio_fd, SNDCTL_DSP_SPEED, &n)==-1)
		return(0);

	n=1;
	if(ioctl(_audio_fd, SNDCTL_DSP_STEREO, &n)==-1)
		return(0);

	if(_qdgdfa_16_bit)
	{
		n=detect_endianess();
		if(ioctl(_audio_fd, SNDCTL_DSP_SETFMT, &n)==-1)
			_qdgdfa_16_bit=0;
	}

	if(!_qdgdfa_16_bit)
	{
		n=AFMT_U8;
		if(ioctl(_audio_fd, SNDCTL_DSP_SETFMT, &n)==-1)
			return(0);
	}

	return(1);
}


static void sound_mix_8(void)
/* mixes the sounds in 8 bit mode */
{
	int n,m;
	int s,ls,rs;

	/* nothing to mix */
	if(_audio_fd==-1) return;

	for(n=0;n < SOUND_BUFFER_SIZE;n+=2)
	{
		ls=rs=0;

		for(m=0;m<=_daemon_seq;m++)
		{
			if(_sounds[m].data_8 && _sounds[m].offset!=-1)
			{
				s=_sounds[m].data_8[_sounds[m].offset++];

				if(_sounds[m].offset==_sounds[m].size)
					_sounds[m].offset=_sounds[m].loop;

				if(_sounds[m].pan!=1)
					ls+=s;
				if(_sounds[m].pan!=-1)
					rs+=s;
			}
		}

		if(ls>255) ls=255;
		if(rs>255) rs=255;

		_sound_buffer_8[n]=(unsigned char) ls;
		_sound_buffer_8[n+1]=(unsigned char) rs;
	}

	write(_audio_fd,_sound_buffer_8,SOUND_BUFFER_SIZE);
}


static void sound_mix_16(void)
/* mixes the sounds in 16 bit mode */
{
	int n,m;
	int s,ls,rs;

	/* nothing to mix */
	if(_audio_fd==-1) return;

	for(n=0;n < SOUND_BUFFER_SIZE;n+=2)
	{
		ls=rs=0;

		for(m=0;m<=_daemon_seq;m++)
		{
			if(_sounds[m].data_16 && _sounds[m].offset!=-1)
			{
				s=_sounds[m].data_16[_sounds[m].offset++];

				if(_sounds[m].offset==_sounds[m].size)
					_sounds[m].offset=_sounds[m].loop;

				if(_sounds[m].pan!=1)
					ls+=s;
				if(_sounds[m].pan!=-1)
					rs+=s;
			}
		}

		if(ls<-32768) ls=-32768;
		if(ls>32767) ls=32767;
		if(rs<-32768) rs=-32768;
		if(rs>32767) rs=32767;

		_sound_buffer_16[n]=(short)ls;
		_sound_buffer_16[n+1]=(short)rs;
	}

	write(_audio_fd,_sound_buffer_16,SOUND_BUFFER_SIZE * 2);
}


static void sound_mix(void)
{
	if(_qdgdfa_16_bit)
		sound_mix_16();
	else
		sound_mix_8();
}


static void sound_reset(void)
{
	int n;

	for(n=0;n < MAX_SOUNDS;n++)
	{
		if(_sounds[n].data!=NULL)
		{
			free(_sounds[n].data);
			_sounds[n].data=NULL;
			_sounds[n].data_8=NULL;
			_sounds[n].data_16=NULL;
		}
	}

	_daemon_seq=-1;
}


static void sound_cmd_load(char * wavfile)
/* load a sound */
{
	FILE * f;
	int size,bits;
	int n,c,c2;

	_daemon_seq++;

	if((f=qdgdfa_load_wav(wavfile, &size, &bits))==NULL)
		return;

	if(!_qdgdfa_16_bit && bits==16) size/=2;
	if(_qdgdfa_16_bit && bits==8) size*=2;

	_sounds[_daemon_seq].data=(unsigned char *) malloc(size);

	/* sound size is stored as # of samples, not bytes */
	if(_qdgdfa_16_bit) size/=2;

	_sounds[_daemon_seq].size=size;
	_sounds[_daemon_seq].loop=-1;
	_sounds[_daemon_seq].offset=-1;
	_sounds[_daemon_seq].pan=0;
	_sounds[_daemon_seq].data_8=(unsigned char *)
		_sounds[_daemon_seq].data;
	_sounds[_daemon_seq].data_16=(short int *)
		_sounds[_daemon_seq].data;


	for(n=0;n<size;n++)
	{
		if(_qdgdfa_16_bit)
		{
			if(bits==8)
			{
				c=fgetc(f)-128;
				c<<=8;
			}
			else
			{
				c2=(unsigned char) fgetc(f);
				c=(signed int) fgetc(f);

				c<<=8;
				c|=c2;
			}

			_sounds[_daemon_seq].data_16[n]=(short int) c;
		}
		else
		{
			if(bits==8)
				c=(unsigned char) fgetc(f);
			else
			{
				fgetc(f);
				c=(signed int) fgetc(f);
				
				c=128+c;
				c&=0xff;
			}

			/* stores as 3/4 of original sound */
			c*=3;
			c>>=2;

			_sounds[_daemon_seq].data_8[n]=(unsigned char) c;
		}
	}

	fclose(f);
}


static int parse_sound_cmd(int fd, char * msg)
/* interpretes messages from the client */
{
	int sound;
	int val;
	int ret;
	char * cmd;
	int n=0;

	ret=1;

	if(_audio_fd==-1) return(ret);

	for(cmd=strtok(msg,";");ret && cmd!=NULL;cmd=strtok(NULL,";"))
	{
		switch(*cmd)
		{
		case SOUND_CMD_EXIT:

			ret=0;
			break;

		case SOUND_CMD_RESET:

			sound_reset();
			break;

		case SOUND_CMD_LOAD:

			sound_cmd_load(cmd+1);
			break;

		case SOUND_CMD_START:

			sound=(int) cmd[1]-'A';
			val=(int) cmd[2]-'A';

			if(_sounds[sound].data)
			{
				_sounds[sound].offset=0;
				_sounds[sound].loop=val?0:-1;
			}

			break;

		case SOUND_CMD_STOP:

			sound=(int) cmd[1]-'A';

			if(_sounds[sound].data)
				_sounds[sound].offset=-1;

			break;

		case SOUND_CMD_SETPAN:

			sound=(int) cmd[1]-'A';
			val=(int) cmd[2]-'A';

			if(_sounds[sound].data)
				_sounds[sound].pan=val;

			break;

		case SOUND_CMD_UNPAUSE:

			if(_audio_fd!=-1) n=1;

			/* falls through */

		case SOUND_CMD_PAUSE:

			ioctl(fd, FIONBIO, &n);
			break;
		}
	}

	return(ret);
}


static volatile void sound_daemon(int _pipe_read)
/* sound server main loop */
{
	int n;

	if(open_device())
	{
		/* non-blocking mode */
		n=1;
		ioctl(_pipe_read, FIONBIO, &n);
	}

	/* no sounds */
	for(n=0;n < MAX_SOUNDS;n++) _sounds[n].data=NULL;

	for(;;)
	{
		if((n=read(_pipe_read,msg,sizeof(msg)))<=0)
		{
			if(errno!=EWOULDBLOCK)
				break;
		}
		else
		{
			msg[n]='\0';

			if(!parse_sound_cmd(_pipe_read,msg))
				break;
		}

		sound_mix();
	}

	if(_audio_fd!=-1)
		close(_audio_fd);

	close(_pipe_read);

	exit(0);
}


/** interface **/

int qdgdfa_load_sound(char * wavfile)
{
	if(!_qdgdfa_sound) return(0);

	msg[0]=SOUND_CMD_LOAD;
	strcpy(&msg[1],wavfile);
	strcat(msg,";");

	write(_command_fd,msg,strlen(msg));

	_program_seq++;
	return(_program_seq);
}


void qdgdfa_play_sound(int snd, int loop)
{
	if(!_qdgdfa_sound) return;

	if(snd<0 || snd>_program_seq)
		return;

	msg[0]=SOUND_CMD_START;
	msg[1]=(char) ('A'+snd);
	msg[2]=(char) ('A'+loop);
	msg[3]=';';

	write(_command_fd,msg,4);
}


void qdgdfa_stop_sound(int snd)
{
	if(!_qdgdfa_sound) return;

	if(snd<0 || snd>_program_seq)
		return;

	msg[0]=SOUND_CMD_STOP;
	msg[1]=(char) ('A'+snd);
	msg[2]=';';

	write(_command_fd,msg,3);
}


void qdgdfa_set_pan(int snd, int pan)
{
	if(!_qdgdfa_sound) return;

	if(snd<0 || snd>_program_seq)
		return;

	msg[0]=SOUND_CMD_SETPAN;
	msg[1]=(char) ('A'+snd);
	msg[2]=(char) ('A'+pan);
	msg[3]=';';

	write(_command_fd,msg,4);
}


void qdgdfa_reset(void)
{
	if(!_qdgdfa_sound) return;

	msg[0]=SOUND_CMD_RESET;
	msg[1]=';';
	write(_command_fd,msg,2);

	_program_seq=-1;
}


void qdgdfa_pause(int p)
{
	if(!_qdgdfa_sound) return;

	msg[0]=p ? SOUND_CMD_PAUSE : SOUND_CMD_UNPAUSE;
	msg[1]=';';
	write(_command_fd,msg,2);

	_program_seq=-1;
}


/** CDROM **/

#ifdef QQ
static void PlayCDROM(void)
{
	struct cdrom_ti trkidx;

	if(_cdrom_fd==-1)
		return;

	memset(&trkidx,'\0',sizeof(trkidx));

	trkidx.cdti_trk0=_cd_track;
	trkidx.cdti_trk1=_cd_track;

	ioctl(_cdrom_fd, CDROMSTOP);
	ioctl(_cdrom_fd, CDROMPLAYTRKIND, &trkidx);

	_cd_playing=1;
}


void CDROM(void)
{
	static int s=0;
	struct cdrom_subchnl subchnl;

	if(_cdrom_fd==-1 || !_cd_track)
		return;

	/* truco para chequear slo una vez de cada ocho */
	if(++s<8)
		return;
	s=0;

	memset(&subchnl,'\0',sizeof(subchnl));

	ioctl(_cdrom_fd, CDROMSUBCHNL, &subchnl);

	if((subchnl.cdsc_audiostatus & 0x2))
		PlayCDROM();
}


void StopCDROM(void)
{
	if(_cdrom_fd==-1 || !_cd_track)
		return;

	ioctl(_cdrom_fd, CDROMSTOP);

	_cd_track=0;
	_cd_playing=0;
}
#endif


/** arranque **/

void qdgdfa_startup(void)
{
	int p[2];

	if(!_qdgdfa_sound) return;

	pipe(p);

	if((_daemon_pid=fork())==0)
	{
		close(p[1]);
		sound_daemon(p[0]);
	}

	close(p[0]);
	_command_fd=p[1];

/*	  if(_music)
		_cdrom_fd=open("/dev/cdrom", O_RDONLY|O_NONBLOCK);
*/
	atexit(qdgdfa_shutdown);
}


void qdgdfa_shutdown(void)
{
	static int _down=0;

	if(!_qdgdfa_sound || _down) return;

	msg[0]=SOUND_CMD_EXIT;
	msg[1]=';';
	write(_command_fd,msg,2);

/*	   StopCDROM(); */

	close(_command_fd);
/*	  close(_cdrom_fd); */

	kill(_daemon_pid, SIGKILL);

	_down=1;
	_qdgdfa_sound=0;
}


void qdgdfa_restart(void)
{
}
